#  boundary_penalty
#
#  2018-08-13  by TAKENAKA, Akio
#  rev. 2020-04-30 

import os.path
import sys
import re

# J: ScatterPenalty に持たせる、保護区のばらつき具合に応じた
# J: ペナルティ計算の実装クラス
# J: 
# J: 散在ペナルティとして、あらたに追加するサイトが既選択サイトと共有する境界線長に
# J: マイナスをつけたものを返す。ペナルティの最大値は0で、サイト追加による外周長増加
# J: が低減されるほどマイナスが大きい = ペナルティが小さい。
# J: 全サイトは共通の外周長を持っていることを想定している。

class BoundaryPenalty:
    """
    This class is an implementation class of scatter penalty calculator
    This class contains length of shared bourder among sites.
    
    As the penalty of scattered sites, it returns length of border shared
    by the newly added site and already selected sites, multiplied by -1.  
    The maximum value is 0.  More border is shared, less will be penalty (more negative).
    It is assumed that all sites are of the same periphery length.
    """
    def __init__(self):

        self.ready = False

        # Dictionary, of which keys are IDs of two sites concatinated 
        # and the values, lengths of boundery shared by the two sites.
        self.boundaries = {}  

        return


    # J: ####################################
    # J: 必要データの読み込みなどが済んでいるか
    def is_ready(self):
        """
        This function returns the value of self.ready (boolean).

        Return
        boolean
            True after loading a data file and get ready to calculate penalty.
            False if not ready.
        """

        return self.ready


    # J: ####################################
    # J: 辺の共有データの読み込み
    def load_data(self, file_name):

        """
        This function loads data of boundary lengths shared by pairs of sites.
        Lines are in the following format: 

        site_ID1[tab]site_ID2[tab]length_of_shared_boundary

        Parameters
        file_name: string
            name of the data file

        Return
        boolean
            True (if successful) or False (if failed)
        """

        if os.path.exists(file_name) is not True:
            print("!! ", file_name, " not found in BoundaryPenalty.load_data")
            return False

        fin = open(file_name, 'rt')

        line = fin.readline() # Header

        while True:
            line = fin.readline()
            if not line:
                break

            data = re.split("\s+", line[:-1])

            if len(data) != 3:
                print("# skip a line in boundary_data_file: ", line[:-1])
                continue

            # IDs of two sites and the length of their boundary
            (site1, site2, len_boundary) = data

            dict_key = self._make_key_from_IDs(site1, site2)
            self.boundaries[dict_key] = float(len_boundary)

        fin.close()

        self.ready = True  # Now ready to evaluate penalties.

        return True

    # J: ####################################
    # J: site の集合に site を追加したときに、既選択サイトと共有する境界長を計算し、
    # J: -1 をかけてペナルティとして返す。より大きなマイナスの値であるほど、サイト追加に
    # J: よる総外周長の増加が少ない。0 が最大値。

    def get_penalty(self, sites, site_to_add):

        """
        This function calculates the change of toral boundary length 
        when a site is added to a set of sites.
        This function calculates the length of border shared by a newly added site 
        and already selected sites, multiplied by -1.  
        More border is shared, less will be penalty (more negative).
        It is assumed that all sites are of the same periphery length.
            
        Parameters
        sites: list
            list of IDs of sites already selected.
        site_to_add: string
            ID of a site to be added to already selected sites.

        Return
        float
            Change of total periphery of selected sites by the inclusion of
            'site_to_add'.  It would be negative if total periphery decreases.
        """

        shared_len_sum = 0
        
        # Test if site_to_add shares boundary with already selected ones
        for site in sites: 

            shared_len = self._get_shared_border_len(site, site_to_add)
            shared_len_sum += shared_len

        return -1 * shared_len_sum

    # J: ####################################
    # J:  ふたつの site が共有する辺の長さを返す
    # J:  共有しないなら 0。

    def _get_shared_border_len(self, id_1, id_2):
        """
        This function returns length of boundary shared by two sites.

        Parameters
        id_1, id_2: string
            IDs of two sites.

        Return
        float
            length of shared boundary. 0 if not shared.
        """

        pair_key = self._make_key_from_IDs(id_1, id_2)

        if pair_key not in self.boundaries:  # No boundary shared
            return 0
        else:
            return self.boundaries[pair_key]


    # J: ####################################
    # J:  ふたつの site ID からキーを作成。

    def _make_key_from_IDs(self, id_1, id_2):
        """
        This function makes a string from two site IDs as a key of 
        a dictionary which contains length of boundary shared by two sites.
        The returned string is insensitive to the order of the two IDs.
        
        Parameters
        id_1, id_2: string
            IDs of two sites.

        Return
        string
            Two IDs concatinated with a space inbetween.
            
        """

        if (id_1 > id_2):
            (id_1, id_2) = (id_2, id_1)

        return "%s %s" % (id_1, id_2)
