#   CoSSPL 
#   (Complemenral Site Selector considering Cost and Penalties)
#
#     2018-12-12  by TAKENAKA, Akio

import random

from . import local_units
from . import site_cost
from . import scatter_penalty

# J:  サイトごとの生物多様性要素の在･不在データに基づく保護区選択
# J:  サイトのコストおよびばらつき度合いによるpenalty も考慮する

class CoSSPL:
    """
    This class is for the conservation area prioritization based on 
    presence-absence data of features to be conserved.
    Greedy algorithm is employed.
    Cost of sites and penalty for scattered area can be taken into account.
    """

    def __init__(self):

        # J: 要素の分布情報を保存する辞書。
        # J: self.data_set[サイト] で「そのサイトに存在する要素のリスト」にアクセス
        # Dictionary to contain distribution of features
        # The keys are site IDs and the values, lists of features 
        # which are present in individual sites.
        self.data_set = {}

        # Dictionary: Keys are Feature IDs and values, abundances of individual features.
        self.feature_occurrence = {}  

        # Dictionary: Keys are Site IDs and values, number of present features.
        self.n_features_in_site = {}  

        # List of IDs of sites which are chosen to start with.
        self.reserved_sites = [] 

        # Parameter to determine how sites with suboptimal conservation effectiveness
        # be considered as candidates.  Larger this parameter, less effective sites
        # may be considered as long as they are less costly and/or contribute to 
        # less scattered conservation area.  If zero, no such upset would be allowed.
        self.max_upset = 0

        # How many times each feature should be covered.
        self.required_occurrence = 1

        # SiteCost class object for cost calculation
        self.site_cost = None

        # Object for calculating penalty for scattered sites
        self.scatter_penalty_obj = None

        # Priority of cost/scatter penalty
        self.consider_cost_first = False   
            # True: Cost considered before scatter pelanty
            # False(default): scatter penalty considered first

        ### Dynamic data members

        # Array of selected sites
        self.selected_sites = []  

        # key: feature ID
        # value: times represented in so far selected sites
        # J: これまでに選択されたサイトの中で、それぞれの要素が何回出現しているか。
        # J: 要素がキー、出現回数が値の辞書
        self.occurrence_in_selected_sites = {}  

    # J: ####################################
    # J:   在データを１件追加する。
    # J:   Usage: obj.add_feature_site_record(要素ID, サイトID)
    # J:   データ追加はかならずこのメソッドで行う

    def add_feature_site_record(self, feature, site):
        """
        This function addd a feature presence data at the given site

        Parameter
        feature: string
            ID of the feature to add
        site: string
            ID of the site where the feature is present
        """

        if ((feature in self.feature_occurrence) == False):  
            # The first time to add this feature
             self.feature_occurrence[feature] = 0

        self.feature_occurrence[feature] += 1

        if ((site in self.n_features_in_site) == False):  
             # The firs time appearance of this site
             self.n_features_in_site[site] = 0

        self.n_features_in_site[site] += 1  

        if ((site in self.data_set) == False):
            # Add a new entry to the dictionary for the given site
            # The value will be a list of feature IDs.
            self.data_set[site] = [] 

        # Add the theature to the list for given site.
        self.data_set[site].append(feature)


    # J: ####################################
    # J: あらかじめ保全するサイトを設定する。
    # J: すでに設定されている場合、上書きする

    def set_reserved_sites(self, sites):
        """
        This function sets sites of pre-reserved area 

        Parameter
        sites:  a list of string
            list of site IDs
        """

        self.reserved_sites = list(sites)


    # J: ####################################
    # J: あらかじめ保全するサイトを追加する

    def add_reserved_sites(self, sites):
        """
        This function appends sites to pre-reserved area.

        Parameter
        sites:  a list of string
            list of site IDs
        """

        for site in sites:
            if site not in self.reserved_sites:
                self.reserved_sites.append(site)


    # J:####################################
    # J: 保全効果最大のサイトから、upset番目まで効果が落ちるサイトまで
    # J: の範囲で、コスト・ペナルティが小さいサイトを選択する。

    def set_max_upset(self, upset):
        """
        This function sets the maximum upset (number of sites behind the top) by sites 
        less effective in covering biodiversity features, but less costly and/or less 
        penalized for scattering

        Parameter
            upset: number of sites behind the top. If 1, no upset occures.
        """

        self.max_upset = upset


    # J:####################################
    # J:  出現回数の最終目標（何回までの出現を求めるか）

    def set_required_occurrence(self, n):
        """
        This function sets required occurrence times fo features.
        
        Parameter
        n: integer
            required occurence time 
        """

        self.required_occurrence = n


    # J:####################################
    # J:  SiteCost クラスのオブジェクトを渡して登録する

    def set_site_cost(self, site_cost_obj):
        """
        This function registers a SiteCost object.
        
        Parameter
            site_cost_obj: an object of SiteCost class
        """

        self.site_cost = site_cost_obj


    # J:####################################
    # J:  ScatterPenalty クラスのオブジェクトを渡して登録する

    def set_scatter_penalty_obj(self, penalry_obj):
        """
        This function registers a ScatterPenalty object.

        Parameter
        penalry_obj: object of ScatterPenalty class
        """

        self.scatter_penalty_obj = penalry_obj


    # J:####################################
    # J:  コストとペナルティのどちらを優先して考慮するか
    # J:  Penalty 優先なら False (default), cost 優先ならTrue に設定

    def set_to_consider_cost_first(self, cost_first = False):
        """
        This fuction sets the priority between the cost and scatter penalty.
        
        Parameter
        cost_first: boolean
            True: Cost considered before scatter pelanty
            False(default): scatter penalty considered first
        """

        self.consider_cost_first = cost_first


    # J:####################################
    # J:  サイト選択を行う。
    # J:  
    # J: 引数 n_sites_to_select で、最大いくつのサイトを選ぶかを与える。
    # J: 指定しない場合 (デフォルト値 0）は制限なしで、各要素の出現回数が条件を
    # J: 満たすまでサイトを選ぶ。
    # J:
    # J: コストやペナルティの評価は、すべて下請けメソッド _pick_effective_site() に
    # J: 任せている
    # J:
    # J: 結果はリストにまとめて返す。リストの要素は、以下のフィールドがタブ区切りで並んだもの。
    # J:
    # J: 何番めの選択サイトか
    # J: サイトID
    # J: 目標出現回数
    # J: 総出現サイト数が目標値以上の要素の数
    # J: そのうち、目標を達成している要素の数

    def select_sites(self, n_sites_to_select = 0):
        """
        This function select sites to cover biodiversity features.
        When maximum number of sites is not given (or 0 is given), it keeps on 
        selecting sites to cover all features as much as the required occurence.
        
        Parameter
        n_sites_to_select: integer
            maximum number of sites to selected.
            If the maxiumum number of sites larger than 0 is given, it stops
            the selection at the time when selected sites reaches the maximum.

        Return
        list of string: 
            each element of the list consits of the following separated by tabs;
                order in the selection process
                site ID
                target occurence of features (increases from 1)
                number of features whose global occurence is equal to or 
                   more than the target occurence
                number of features fulfilling the target occurence
        """

        # J: 初期保全サイトを作業用リストにコピー。ここから順次とっていく
        # Make a copy of pre-reserved sites
        reserved = list(self.reserved_sites)  

        # J: 事前保護サイトがあって、これが分布データに含まれないサイトの場合への対応
        # Just in case a pre-reserved site was not included in the local units data file.
        if len(reserved) > 0 :  
            # J: 事前保護サイトが含まれないならそのサイトを登録
            for r_site in reserved:  
                self._add_site(r_site)

        # J: 全サイトのリスト
        sites  = self._get_all_sites()  
        n_sites = len(sites)

        # J: 選択サイト数の指定がなければ全サイト。(総サイト数以上は選べない)
        # If maximum number of sites to be selected is not given, it is 
        # assumed to be the number of all sites.

        if n_sites_to_select == 0:
            # J: 全サイトまで選べるように。
            n_sites_to_select = n_sites  

        # J: 全サイト（事前保全サイトはあとで削除）
        remaining_sites = sites

        # Shuffled to add stochasticity in the order of selection
        random.shuffle(remaining_sites) 

        # J: 「N箇所以上に出現する要素類数」の対応表作り（N がキー、要素数が値の辞書）
        # Dictionary. 
        #  Keys will be N, occurrence time of individual features,
        #  value, number of features of which occurence time is N or more.

        n_features_satisfy_given_occurrence = self._calc_occurrence_distribution()

        self.selected_sites = []             # Will have selected sites
        self.effects        = []             # Effects of selected sites

        # J: 選択サイト内の要素の数（辞書）を初期化
        self.occurrence_in_selected_sites = {} 

        features = self._get_all_features()

        # J: 各要素ごとにゼロで初期化
        for feature in features:       
            self.occurrence_in_selected_sites[feature] = 0

        # J: 何サイト以上での出現を要求するか。
        # Requirement of occurrence time.
        repl = 1;    # Starts from 1.

        # J: 条件を満たし得る要素の数
        # Number of features that may achieve the occurrence requirement
        n_features_to_be_covered = n_features_satisfy_given_occurrence[repl]


        for i in range(n_sites_to_select): # increase number of sites

            # J: 初期保全サイトがあれば、そこからひとつ取り出す。
            # Pre-reserved sites have priority
            if len(reserved) > 0:  
                # J: 先頭から取り出す
                selected_site = reserved.pop(0) 

            # J: 初期保全サイトの指定がない、ないしはすべて取り出した場合
            # J: 残りのサイトの中でもっとも効果的なサイトを選ぶ。
            # If there is no pre-reserved sites, or all has been popped,
            # select the most effective site among the remaining sites.
            else: 

                selected_site = self._pick_effective_site(repl, remaining_sites)
                # print("selectted %s : NEW" % selected_site)

                if selected_site is None:
                    print("No more site available to improve conservation effect")
                    return

            # J: 選択されたサイトを、選択サイトリストに加える。
            # Add the selected site to the protected area.
            self._add_site_to_protected_area(selected_site)

            # J: 未選択サイトのリストから、新たに選んだサイトを削除。
            # J: (事前保全サイトもここで）
            # Remove the selected sites from the pool of sites
            remaining_sites.remove(selected_site)

            # J: 選択されたサイトにより、必要出現回数を満たした要素の数。
            # Number of features covered required times by adding the selected site.
            n_features_covered = self._get_covered_features_num(repl)
            # print(" Added ", selected_site, "- n_features_covered", n_features_covered)

            # J: 結果を保存。
            # Store the selected site and its effect.
            result =  "%d\t%s\t%d\t%d\t%d" % \
                    (i + 1, selected_site, repl, n_features_to_be_covered, n_features_covered)

            self.effects.append(result)

            # J: 現在の必要出現回数 repl が満たされた場合はreplを増やす。
            # J: すでに最終目標まで達成されているなら終了（ただし事前保護サイトは全部足す）
            # If all features to be covered are already covered, increase the
            # requirement. 
            # If the requirement has reached the final goal, the selection 
            # is terminated.

            if (n_features_covered == n_features_to_be_covered):

                # J: すでに最終目標を達成している場合
                # If the final goal is acheved...
               
                if repl >= self.required_occurrence: 

                    # J: 事前保護区サイトが残っているならそれを保護区に足す
                    # In case pre-reserved sites are remaining
                    # (i.e., they are more than enough to achieve target)

                    # J: 初期保全サイトがまだあれば、
                    while len(reserved) > 0:     
                        # J: 先頭から取り出す
                        site = reserved.pop(0)   
                        # J: 保護区に足す
                        self._add_site_to_protected_area(site)  


                    # J: for ループを抜ける
                    break


                # J: まだ最終目標を達成していない場合（repl が self.required_occurrence 未満）
                # When required_occurrence is not yet met.

                repl += 1 # Increase the current requirement

                # J: その要求条件を満たせる（存在サイト数が要求回数以上）の要素はいくつか？
                # How many features can satisfy the requiement?

                if repl not in n_features_satisfy_given_occurrence:
                    # J: repl 回以上出現する要素はない場合
                    # No features can.
                    n_features_to_be_covered = 0

                else:
                    # J: repl 回以上出現する要素がある場合
                    # Some features still can.
                    n_features_to_be_covered = n_features_satisfy_given_occurrence[repl]


                # J: これ以上、カバーするべき要素はないのであれば終了。
                # If no more features are to be covered...
                if n_features_to_be_covered == 0:

                    # J: ただし、まだ事前保護区のサイトが残っているなら、それを保護区に足す作業を済ませる
                    # In case pre-reserved sites are remaining

                    if len(reserved) > 0:  
                        for site in reserved:
                            self._add_site_to_protected_area(site)

                    # As no more features are to be covered, get out of the loop
                    break


    # J: ##################
    # J: 選ばれたサイトを、選ばれた順に並べたタプルを返す。

    def get_selected_sites(self):
        """
        This function returns IDs of selected sites.
        
        Return
        Tuple of strings:
            tuple of IDs of selected sites in the order of selection
            (sites selected later appear later)
        """

        return tuple(self.selected_sites)


    ######  Following functions are to be used only within this class ########

    # J: ####################################
    # J: データセットにサイトを追加する。
    # J: すでにデータセットに含まれるサイトなら何もしない。
    # J: 要素が存在しないサイトなら、そのサイトの存在を記録する。
    # J:
    # J:  Usage: obj._add_site( サイトID)

    def _add_site(self, site):
        """
        This function adds a site which is not registered yet.
        If the given site is already registered, this function does nothing.
        
        Parameter
        site: string
            ID of a site to add
        """

        # If not registered yet...

        if ((site in self.n_features_in_site) == False):  
             self.n_features_in_site[site] = 0 

        if ((site in self.data_set) == False):
            self.data_set[site] = [] 


    # J: ######################
    # J: 「N箇所以上に出現する要素数」の対応表作り（N がキー、要素数が値の辞書）
    # J: 
    # J:  Nの最大値を Nmax とすると、N が 1 から Nmax のすべてについて要素が定義
    # J:  される。

    def _calc_occurrence_distribution(self):
        """
        This function makes a summary table of occurrence times of features.

        Return
        dictionary:
            Keys will be N, occurrence time of individual features,
            value, number of features of which occurence time is N or more.
        """

        num_features_satisfy_given_occurrence = {}

        features = self._get_all_features()

        for feature in features: 

            # time of occurence of the fieatre
            times = self.feature_occurrence[feature];

            # J: N 回出現するなら, １ケ所以上に出現する要素数、２ケ所以上に出現する要素数、
            # J: …、Nケ所以上に出現する要素数、をそれぞれ１づつ増やす。

            for i in range(1, times + 1):
                if (i in num_features_satisfy_given_occurrence) == False:
                    num_features_satisfy_given_occurrence[i] = 1
                else:
                    num_features_satisfy_given_occurrence[i] += 1

        # J: 辞書を返す
        return num_features_satisfy_given_occurrence  


    # J: ###################
    # J: 全要素のリスト

    def _get_all_features(self):
        """
        This function returns a list of IDs(string) of all features 
        
        Return
        list of IDs(string) of all features
        """

        features =  list(self.feature_occurrence.keys())
        
        return features


    # J: ###################
    # J: 全サイトのリスト

    def _get_all_sites(self):
        """
        This function returns a list of IDs(string) of all sites 
        
        Return
        list of IDs(string) of all sites
        """

        sites  = list( self.n_features_in_site.keys() )

        return sites


    # J: ###################
    # J: これまでの選択サイトにより、指定した必要出現回数を満たしている要素の数を返す。

    def _get_covered_features_num(self, n):
        """
        This function returns number of features which has been covered 
        n or more times by the so far selected sites.

        Parameter
        n: times of occurrence (integer)

        Return
        number of features covered (integer)
        """

        n_features_covered = 0

        features = self._get_all_features()

        for feature in features :

            if self.occurrence_in_selected_sites[feature] >= n:
                n_features_covered += 1

        return n_features_covered


    # J: ###########################################
    # J: 選択サイトをひとつ加える。
    
    def _add_site_to_protected_area(self, site):
        """
        This function adds a given site to the set of protected sites
        
        Parameter
        site: ID (string) of a site to be added
        """

        self.selected_sites.append(site)

        # Update times of occurence in protected sites
        # for all features in the added site

        for feature in self.data_set[site]:
            self.occurrence_in_selected_sites[feature] += 1


    # J: #############################
    # J:  必要出現回数および未選択サイトの配列を受けとり、
    # J:  要素をカバーする効果の高いサイトをひとつ選んで返す。
    # J:  効果のあるサイトがなければ None を返す。

    def _pick_effective_site(self, n_occurrence, sites):
        """
        This function returns a site among given sites most effective in
        covering features satisfying required occurrence time when added to the
        current protected area.
        If all the given sites has no effect, None is returned.
        If maximum upset larger than 1 is set, less effective but with
        less cost or penalty may be returned.
        
        Parameter
        n_occurrence: required occurrence time (integer)

        Return
        ID of the most effective site (string)
        None if no site has any effect
        """

        # J: n_candidates は、保全効果トップから、最大逆転可能順位のサイトまでのサイト数
        # n_candidates is a number of runners-up which can upset the most effective site 
        # if they are less costly or less scattered.

        if self.max_upset + 1 < len(sites):
            n_candidates = self.max_upset + 1 
        else:
            # Note that it cannot be more than the number of given sites.
            n_candidates = len(sites)  

        # J: 候補サイトそれぞれを選んだ場合の、目標達成要素数の増分
        # J: サイトIDがキー、要素数の増分が値の辞書
        # The effect of adding each of the given sites.
        # Keys are site IDs and the values, number of features that newly clear
        # the required times of occurrence.

        effects = self._calc_n_additional_features_duly_represented(n_occurrence, sites)

        # Get a list of effects sorted in descending order.
        effect_vals_ = effects.values() 
        effect_vals  = list(effect_vals_)
        effect_vals.sort(reverse = True)

        max_effect = effect_vals[0]  # The first (0'th) element is the maximum
        
        if max_effect == 0: # No improvement by any of the given sites.
            return None

        # J: 保全効果トップから n_candidates めのサイトの保全効果( effect_vals[n_candidates - 1] )
        # J: 以上の効果があるサイトが候補となる。
        # J: ただし効果ゼロであってはいけないので、非ゼロの値をうしろ（小さい方）から探す。
        # The value of n_canditates'th element if the threshold to be the runners-up
        # Note that it should be larger than 0.

        for i in range(n_candidates - 1, -1, -1):
            if effect_vals[i] > 0:
                threshold_val = effect_vals[i] 
                break

        # List to store candidate sites

        # Sites of which effect is threshold_val or larger
        candidates = list()      

        # Sites of which effect is the maximum (can be two or more)
        top_candidates = list()  

        for site in sites:
            if effects[site] >= threshold_val:  # Clear threshold to be runners-up
                candidates.append(site)

            if effects[site] == max_effect:    # Of the maximu effect
                top_candidates.append(site)

        # If cost and scatter penalty are not considered, choose one randomly
        # from top_candidates
        if self.scatter_penalty_obj is None and self.site_cost is None:
            return random.choice(top_candidates)

        # In the following, cost and/or scatter penalty is considered.

        penalties  = dict() # Keys will be site IDs, values will be penalties
        costs  = dict()     # Keys will be site IDs, values will be costs

        for site in candidates:
            penalties[site] = self._get_penalty(self.selected_sites, site)
            costs[site] = self._get_cost(site) 

        # J:コストとペナルティの優先順位
        # J: 基本は、周辺長ペナルティ優先。選択の初期はどのみちどのサイトでも
        # J: 既選択サイトと接する可能性は低く、コストで選択される。途中から、
        # J: 既選択サイトを核として隣接するサイトが選ばれていく。
        # J: コスト優先に設定すると、最後まで周辺長ペナルティの効果が小さいまま。
        # Priority of cost and scatter penalty

        if self.consider_cost_first == True:
            first_ = costs
            second_ = penalties
        else:
            first_ = penalties
            second_ = costs

        # J: first_ の最小値（コストないしペナルティの）
        # Minimum cost or penalty, whichever of higher priority
        min_1 = min(first_.values())  

        # J:  first_ が最小値のサイト群
        # Sites with the minimum cost or penalty, including ties.
        preferable_sites = list()  

        # Penalty or cost,  whichever of lower priority,
        #  of the sites in preferable_sites
        second_vals = list()

        for site in candidates:
            if first_[site] == min_1:
                preferable_sites.append(site) 
                second_vals.append(second_[site])

        # Minimum of penalty or cost, whichever of lower priority,
        # among sites in preferable sites.
        min_2 = min(second_vals)  

        # List of sites with least cost or penalty, whichever is of higher priority,
        # and also least in other, among the least cost/penalty sites.
        # The finalists may include sites not most effective in covering features.

        finalist_sites = list() 

        for site in preferable_sites:
            if second_[site] == min_2:
                finalist_sites.append(site)

        # Choose one of the finalits
        return random.choice(finalist_sites)  


    # J: ####################
    # J: ひとつのサイト site を保護区に加えることによるコストの増加。

    def _get_cost(self, site):
        """
        This function returns cost increment by adding the given site 
        to the present protected area. If SiteCost object is not 
        registered, it returns 0.
        
        Parameter
        site: string
            ID of a site
        Return
        increment of cost (float)
            If no cost object is set, 0 is returned.
        """

        cost = 0

        if self.site_cost is not None:
            cost = self.site_cost.get_cost_of_site(site)

        return cost


    # J: ####################
    # J: サイトのリスト sites に、ひとつのサイト site_to_add を合わせた site のセット
    # J: について、ScatterPenalty クラスのオブジェクトにペナルティを計算させて結果を返す。

    def _get_penalty(self, sites, site_to_add):
        """
        This function returns scatter penalty of adding a given site
        to a set of sites. If ScatterPenalty object is not registered, 
        it returns 0.
        
        Parameter
        sites: list of string
            list of site IDs
        site_to_add: string
            ID of a site to add to 'sites'
        
        Return
        float
            scatter penalty of adding 'site_to_add' to 'sites'
            If no ScatterPenalty object is set, 0 is returned.
        """

        if self.scatter_penalty_obj is None:
            return 0

        penalty = self.scatter_penalty_obj.get_penalty(sites, site_to_add)

        return penalty


    # J: ###########################################
    # J: 必要出現回数と、サイトIDの配列（レファレンス）を受けとり、
    # J:  それぞれのサイトを選択に加えた場合、何要素があらたに出現回数の条件
    # J:  を満たすかを計算して返す。
    # J:  返値は辞書。サイトIDがキー、条件を満たす要素数が値

    def _calc_n_additional_features_duly_represented(self, n_occurrence, candidate_sites):
        """
        This function 

        Parameter
        n_occurrence: required occurence time (integer)
        candidate_sites: list of IDs of sites (strings)

        Return
        Dictionary:
            The keys will be site IDs given as 'candidate_sites' 
            and the values, number of features that would newly satisfy 
            the required occurece time by adding the site.
        """

        # J: サイトIDがキー、
        # J: そのサイトを加えるとあらたに出現回数条件を満たす要素数が値の辞書

        # Create a dictionary.
        n_additional_features_duly_represented = {}

        for site in candidate_sites:

            # The keys of the dictionary will be site IDs given 
            # as 'candidate_sites' and the values, number of features 
            # that would newly satisfy the required occurece time 
            #  by adding the site.
            n_additional_features_duly_represented[site] = 0

            for feature in self.data_set[site] :  # For features appearing in the site

                # J: このサイトを選択に加えると出現回数がちょうど条件回数となるなら、
                # J: あらたに条件を満たす要素数に 1 を加える。
                # If the occurrence time so far is just n_occurrence - 1,
                # addition of this site makes the feature clarify the requirement.
                if self.occurrence_in_selected_sites[feature] == n_occurrence - 1:
                    n_additional_features_duly_represented[site] += 1

        return n_additional_features_duly_represented
