#  distance_to_nearest
#
#  2020-05-09  by TAKENAKA, Akio

import os.path
import sys
import re

# J: ScatterPenalty に持たせる、保護区のばらつき具合に応じた
# J: ペナルティ計算の実装クラス
# J: 
# J: あらたに追加するサイトと、既に選択したサイトとの最短距離を
# J: ペナルティとする。データファイルで、各サイトの位置情報(x, y)を読み込む

class DistanceToNearest:
    """
    This class is an implementation class of scatter penalty calculator
    
    """
    def __init__(self):

        self.ready = False

        # Dictionary, of which keys are site IDs 
        # and the value, tuple of x-y coordinates.
        self.coordinate = {}  

        return

    # J: ####################################
    # J: 必要データの読み込みなどが済んでいるか
    def is_ready(self):
        """
        This function returns the value of self.ready (boolean).

        Return
        boolean
            True after loading a data file and get ready to calculate penalty.
            False if not ready.
        """

        return self.ready


    # J: ####################################
    # J: X_Y 座標の読み込み
    def load_data(self, file_name):

        """
        This function loads data of x-y coordinates of sites

        Each line is in the following format: 

        site_ID[tab]x[tab]y

        Parameters
        file_name: string
            name of the data file

        Return
        boolean
            True (if successful) or False (if failed)
        """

        if os.path.exists(file_name) is not True:
            print("!! ", file_name, " not found in DistanceToNearest.load_data")
            return False

        fin = open(file_name, 'rt')

        line = fin.readline() # Header

        while True:
            line = fin.readline()
            if not line:
                break

            data = re.split("\s+", line[:-1])

            if len(data) != 3:
                print("# skip a line in x_y_coordinates_file: ", line[:-1])
                continue

            # site ID and x-y coordinate
            (site_ID, x, y) = data
            self.coordinate[site_ID] = (float(x), float(y))

        fin.close()

        self.ready = True  # Now ready to evaluate penalties.

        return True


    # J: ####################################
    # J: site の集合中、追加する site ともっとも近いものまでの距離を計算し、ペナルティとして返す。
    # J: 既存サイトが空集合なら0を返す。

    def get_penalty(self, sites, site_to_add):

        """
        This function calculates the distance from a site to the nearest 
        of the set of sites.
        
        Parameters
        sites: list
            list of IDs of sites already selected.
        site_to_ad: string
            ID of a site to be added to already selected sites.

        Return
        float
            Distance from 'site_to_add' to the nearest among 'sites'.
        """

        if len(sites) == 0:
            return 0

        min_distance = None  # initial value

        for site in sites: 
            distance = self._get_distance(site_to_add, site)
            if min_distance is None:
                min_distance = distance
            elif min_distance > distance:
                min_distance = distance

        return min_distance


    # J: ####################################
    # J:  ふたつの site のあいだの距離を返す

    def _get_distance(self, id_1, id_2):
        """
        This function returns the distance between two sites.

        Parameters
        id_1, id_2: string
            IDs of two sites.

        Return
        float
            Distance between the two sites.
        """

        coordinate_1 = self.coordinate[id_1]
        coordinate_2 = self.coordinate[id_2]

        distance = ((coordinate_1[0] -  coordinate_2[0]) ** 2
                    + (coordinate_1[1] -  coordinate_2[1]) ** 2) ** 0.5
        
        return distance

