#  LocalUnits
#
# 2020-01-25  by TAKENAKA, Akio
# 
# rev 2023-09-22 b TaKENAKA, Akio
# self.features と self.sites を list から set に変更し、
# 大量データの読み込み時に著しく遅くなるのを抑制

import os.path
import sys
import re
import random

CHR_CODE = 'utf-8'

# J: 多要素・多サイトで、サイトごとにまとめた要素の局所単位(local unit）の
# J: 集合を扱うクラスの定義。
# J: ある要素があるサイトに存在する場合、その要素は 0 より大きい評価値を持つと想定。
# J: 評価値は整数、実数どちらでも可。

class LocalUnits:
    """
    This class manages a collection of local units of multiple features 
    distributed in multiple sites.
    If a feature is present in a site, it has a value larger than 0.
    The value may be integer or float.
    """

    def __init__(self):

        # J: 要素の存在情報を保存する辞書。
        # Dictionaries to store info of local units of features. For efficiency, 
        # values of local units can be accessed in two ways: by feature and site, 
        # and by site and feature.

        self.by_feature = {}  # The keys will be feature IDs and tha vlues, dictionaries. 
                              # The keys of the nested dictionary will be site IDs and 
                              # the values, the value of local units in individual sites.
        self.by_site = {}     # The keys will be site IDs and tha vlues, dictionaries. 
                              # The keys of the nested dictionary will be feature IDs and 
                              # the values, the value of local units of individual features.


        self.highest_value = 0  # The highest value of local units. 

        # J: feature がキー、各feature のlocal unit の評価値のリストが値
        # The keys will be feature IDs and the values, 
        # lists of values of features' local units.
        self.values_of_feature = dict()  

        # J: feature がキー、値は辞書（各local unit の値がキー、存在するサイトのリストが値）
        # The keys will be feature IDs and the values are dictionaries.
        # The dictionaries' keys are local unit values and values, lists of IDs of sites
        # which include the local units of the feature of the given values.

        self.sites_with_feature_of_value = dict()


        # Obsolete
        # self.features = list()
        # self.sites = list()
        
        self.features = set()
        self.sites = set()
        
        return


    # J: ####################################
    # J: オブジェクトの複製。深いコピーを作成して返す。

    def dup(self):
        """
        This function makes a deep copy of itself.
        
        Return
            A deep copy of itself (an object of LocalUnits)
        """

        new_obj = LocalUnits()

        for feature in self.features:

            sites = self.get_sites_with_feature(feature)

            for site in sites:
                val = self.get_value_of_feature_in_site(feature, site)
                new_obj.add_local_unit(feature, site, val)

        return new_obj


    # J: ####################################
    # J: 自身から、指定要素のデータを除いて新規に作成したLocalUnits オブジェクト
    # J: を返す。除く要素のリストを引数で渡す

    def dup_without_given_features(self, features_to_eliminate):
        """
        This function returns a deep copy of itself without given features.
        
        Parameters
            features_to_eliminate: list of string
                a list of feature IDs to eliminate
        
        Return
            A deep copy of itself without features in features_to_eliminate
        """

        new_obj = LocalUnits()

        for feature in self.features:

            # Skip if the feature is to be eliminated
            if feature in features_to_eliminate: 
                continue

            sites = self.get_sites_with_feature(feature)

            for site in sites:
                val = self.get_value_of_feature_in_site(feature, site)
                new_obj.add_local_unit(feature, site, val)

        return new_obj



    # J: ####################################
    # J:  local_unit 評価データファイルを読み込む
    # J: 
    # J: データはタブ区切り。
    # J:  行がサイト、列が要素（feature）に対応
    # J:  １行目は要素ID が並んだヘッダ行。
    # J:  二行目からは、サイトIDおよび各要素のlocal unit の value（評価値）
    # J:  評価レベルは自然数。ゼロ以下の場合は登録しない
    # J:  少数部分は切り捨てられる。
    # J:  読み込みに成功したら True, ファイルがなく読み込めなければ Falseを返す。

    def load_from_file(self, file_name):
        """
        This function loads data from a text file
        The header line consists of feature IDs.
        The following lines consists with a site ID and values of local units 
            of all features in the site. Data are separated by tabs.
        The values of each local unit should be integer larger than 0.
        Decimal parts are ignored.
        If the value is 0 or less, the featuret is interpreted as absent
        
        
        Parameters
            file_name: string
                name of a file to load from
        
        Return
            boolean:
                True if successfully loaded and False, if failed.
        """

        if os.path.exists(file_name) is not True:
            print("!! ", file_name, " not found in LocalUnits.load_from_file")
            return False

        fin = open(file_name, 'rt', encoding = CHR_CODE)

        header = fin.readline()

        items = re.split('\s+', header[:-1])  #  list of feature IDs
        items = items[1:]  # remove the first one (not a feature ID)

        # self.features = items  # list of feature IDs.
        feature_IDs = items  # list of feature IDs.

        while True:

            line = fin.readline()  # read each line
            if not line:
                break

            items = re.split('\s+', line[:-1])  # tap separated, but accepts spaces.

            site = items.pop(0)  # The first field is a site ID.


            for feature, item in zip(feature_IDs, items):
                # float values are converted to int.
                value = int(float(item)) 
                # non-positive values are ignored (interpreted as absence of the feature)
                if (value > 0):                 # Stored if larger than 0.
                    self.add_local_unit(feature, site, value)

        fin.close()

        return True


    # J: #########################
    # J: local unit データを１件追加する。
    # J: local unit の登録は、内部的にもすべてこのメソッドを経由して行う。

    def add_local_unit(self, feature, site, value):
        """
        This function is used to store a data of a local unit.
        It is called both from inside and outside of this class.

        Parameters
            feature: string
                feature ID of a local unit
            site: string
                site ID of the local unit
            value: integer
                The value of the local unit
        """

        # If the feature has not yet been stored...

        if feature not in self.features:
            # self.features.append(feature)
            self.features.add(feature)

        if feature not in self.by_feature: 
            self.by_feature[feature] = {}

        self.by_feature[feature][site] = value

        # If the site has not yet been stored...

        if site not in self.sites:
            # self.sites.append(site)   
            self.sites.add(site)   

        if site not in self.by_site:
            self.by_site[site] = {}

        self.by_site[site][feature] = value

        # Update maximum value

        if value > self.highest_value:
            self.highest_value = value

        # Update dictionaries and lists,

        if feature not in self.values_of_feature:
            self.values_of_feature[feature] = []

        self.values_of_feature[feature].append(value)

        if feature not in self.sites_with_feature_of_value:
            self.sites_with_feature_of_value[feature] = {}

        if value not in self.sites_with_feature_of_value[feature]:
            self.sites_with_feature_of_value[feature][value] = []
        
        self.sites_with_feature_of_value[feature][value].append(site)


    # J: #########################
    # J: local unit データを１件削除する。
    # J: 削除に成功したら True, 
    # J: 対応するデータがなく削除できなかったら False を返す。
    def remove_local_unit(self, site, feature):
        """
        This function removes a local unit of a given feature in a given site.

        Parameters
            site: string
                site ID of the local unit to remove
            feature: string
                feature ID of a local unit to remove

        Return
            boolean:
                True if successfully removed and False, if failed.
        """

        if site not in self.by_feature[feature]:   # The given site is not found
            return False

        value = self.by_feature[feature][site]  

        # Remove from a dictionary by feature
        del self.by_feature[feature][site]

        # J: このサイトが、feature の唯一のサイトであった場合
        if not self.by_feature[feature]:  # If no more local unit of this feature remains
            del self.by_feature[feature]  # remove this feature from the keys of the dictionary


        # Remove from a dictionary by site
        del self.by_site[site][feature]

        # J: この要素が、site の唯一の要素であった場合
        if not self.by_site[site]:  # If no more local unit in this feature remains
            del self.by_site[site]  # remove this site from the keys of the dictionary


        # J: feature の評価値リストから除く
        self.values_of_feature[feature].remove(value)

        # J: このfeature の値リストが空になったら、辞書の項目ごと削除
        if not self.values_of_feature[feature]:
            del self.values_of_feature[feature]


        self.sites_with_feature_of_value[feature][value].remove(site)

        # J: リストが空になったら、辞書の項目ごと削除
        if not self.sites_with_feature_of_value[feature][value]:
            del self.sites_with_feature_of_value[feature][value]

            if not self.sites_with_feature_of_value[feature]:
                del self.sites_with_feature_of_value[feature]

        return True


    # J: ####################################
    # J:  local_unit のもっとも高い評価値（value）を返す。

    def get_highest_value(self):
        """
        This function returns the highest value of all local units.

        Return
            integer:
                The highest value

        """

        return  self.highest_value


    # J: ####################################
    # J: 指定サイト（ひとつ）に存在する要素のリストを返す

    def get_features_in_site(self, site):
        """
        This function returns a list of features in a given site.
        
        Parameter
            site: string
                site ID

        Return
            list of strings:
                list of feature IDs in the given site.
        """

        if site not in self.by_site:
            return []  # return empty list if the given site is not found.
        else:
            return list(self.by_site[site].keys())


    # J: ####################################
    # J: 全要素のリストを返す。

    def get_features(self):
        """
        This function returns IDs of all features.
        
        Return
            list of strings:
                list of all feature IDs
        """

        return list(self.features)   # set を list に変換


    # J: ####################################
    # J: 全サイトのリストを返す。

    def get_sites(self):
        """
        This function returns IDs of all sites.
        
        Return
            list of strings:
                list of all site IDs
        """

        return list(self.sites)  # set を list に変換


    # J: ####################################
    # J: 指定した要素が存在するサイトのリストを返す

    def get_sites_with_feature(self, feature):
        """
        This function returns IDs of sites with a given feature.
        
        Parameters
            feature: string
                feature ID

        Return
            list of strings:
                list of all site IDs
        """

        if feature not in self.by_feature: # When the given feature is not present
            return [] 

        return list(self.by_feature[feature])


    # J: ####################################
    # J: 指定した要素の、指定した value のlocal_unit が存在するサイトを返す(リスト)

    def get_sites_with_feature_of_given_value(self, feature, value):
        """
        This function returns IDs of sites with a local unit of a given feature 
        and of a given value.
        
        Parameters
            feature: string
                feature ID
            value: float
                value of a local unit

        Return
            list of strings:
                list of IDs of sites with the local unit of the given feature 
                of given value
        """

        if feature not in self.by_feature:  # When the feature is not present
            return [] 

        if value not in self.sites_with_feature_of_value[feature]:
            return []   # If no local unit of this feature has the given value

        return list(self.sites_with_feature_of_value[feature][value])


    # J: ############################
    # J: 指定した要素の全local_unitの value の取得。
    # J: 返値は value のリスト（順不同）
    # J: feature の local_unit が存在しない場合は空のリストを返す。

    def get_values_of_feature(self, feature):
        """
        This function returns values of all local units of a given feature
        
        Parameter
            feature: string
                feature ID
        
        Return
            list of integers:
                list of values of the local units of the given feature.
                The values are not sorted.
        """

        if feature not in self.values_of_feature:
            return []

        return list(self.values_of_feature[feature])


    # J: ############################
    # J: 指定した要素の指定サイトのlocal_unitの value の情報の取得
    # J: feature がこのオブジェクトに存在しない場合は例外を発生。
    # J: 指定サイトにfeature の local_unit が存在しない場合は0 を返す

    def get_value_of_feature_in_site(self, feature, site):
        """
        This function returns value of the local units of a given feature
        
        Parameters
            feature: string
                feature ID
            site: string
                site ID
        
        Return
            integer:
                the value of the local unit of the given feature in the given site.
                It there is no local unit of the feature in the given site, 0 is returned.
                If the feature is nowhere, exception is raised.
        """

        if feature not in self.by_feature:
            print("feature", feature, "not in this LocalUnits object")
            raise Exception()

        if site not in self.by_feature[feature]:
            return 0

        return self.by_feature[feature][site]


    # J: ############################
    # J: 指定した要素の指定サイト群内のlocal_unitの value の情報の取得
    # J: 返値は value のリスト（順不同）
    # J: 指定サイトにfeature の local_unit が存在しない場合は空のリストを返す。

    def get_values_of_feature_in_sites(self, feature, sites):
        """
        This function returns values of the local units of a given feature 
        in given sites. The values are not sorted.
        
        Parameters
            feature: string
                feature ID
        
        Return
            list of integer:
                list of values of local units of the given feature 
                in the given sites.
        """
        
        values = []

        for site in sites:

            if site in self.by_feature[feature]:
                value = self.by_feature[feature][site]
                values.append(value)

        return values


    # J: ############################
    # J: sites で与えられたサイトの配列に、指定feature の value の大きいほうから
    # J: 何番目のlocal_unit まで含まれているか

    def howmany_tops_included(self, feature, sites):
        """
        This function returns how many top local units of a given feature
        is included in the given sites.
        
        Parameters
            feature: string
                feature ID
            sites: list of strings
                site IDs
        Return
            integer:
                Number of top local units of the given feature
                included in the given sites.
        """

        # J: feature のlocal_unit の value の配列（降順）
        # List of values of the local units of the given feature
        # Values are sorted in descending order.
        values_of_feature = self.get_values_of_feature(feature);
        values_of_feature.sort()
        values_of_feature.reverse()

        # J: 指定サイトに限った value の配列
        # List of values of the local units of the given feature in the given sites.
        # Values are sorted in descending order.
        values_in_sites = self.get_values_of_feature_in_sites(feature, sites)
        values_in_sites.sort()
        values_in_sites.reverse()

        n_all_units = len(values_in_sites)  # Number of local units of the feature
        top_n = 0

        # J:value の配列を先頭（上位）から比較
        # Compare the values of the above two lists.
        for i in range(0, n_all_units):   

            if values_of_feature[i] > values_in_sites[i]:    
                # J: 指定サイト集合中の valueのほうが小さい
                # J: → 上から (i + 1)個が含まれていないので、i を返す。
                # (i+1)'th value in the list of given sites is smaller than that in the list of 
                #  all local units, which means that top-i is included but top-(i+1) is not.
                # Thus, 'i' is returned.
                return i       

        # J: すべて含まれている場合
        # All unts included, which means that top-n_all_units is included.
        return n_all_units


    # J: ############################
    # J: sites で与えられたサイトのリスト中、指定feature の target (top からの個数）達成に
    # J: 貢献しているサイトを ピックアップしてリストとして返す。
    # J: 同点サイトのうち一部が貢献している場合、必要数のサイトがランダムに選ばれる。
    # J: target が達成されていない場合は、与えられたサイトで実現されている top-n に
    # J: 貢献しているサイトのリストを返す（したがってリストのサイズはtarget より小さい）

    def get_contributing_sites(self, feature, sites, target):
        """
        This function finds sites contributing to the achieve target 
        of the given feature from the given list of sites.

        Parameters
            feature: string
                ID of feature to examine.
            sites: list of string
                IDs of sites to examine.
            target: integer
                number of top value local units to be covered.

        Return
            list of string:
                list of IDs of sites contributing to achieve the target
                (top-'target' included).
                When only part of ties is needed to achieve the target,
                neccessary number of sites are selected randomly.
                When the given sites are not sufficient to achieve the 
                target, the target is reduced.
        """

        # J: 実現されている保全目標
        realized_target = self.howmany_tops_included(feature, sites)

        # J: given の target を達成していないなら、達成できているところまで。
        # If target is not fully achieved, upto the currently achieved.
        if realized_target < target:  
            target = realized_target  

        # J: feature のlocal_unit の value の配列（降順）
        # List of values of the local units of the given feature
        # Values are sorted in descending order.
        values_of_feature = self.get_values_of_feature(feature);
        values_of_feature.sort()
        values_of_feature.reverse()

        # J: 目標達成に、各valueが何個必要か
        # Number of local units of given value (key) to achieve the target
        needed_units = dict()

        count = 0

        for val in values_of_feature:
            if not val in needed_units:
                needed_units[val] = 1
            else:
                needed_units[val] += 1

            count += 1
            if count >= target:
                break

        # J: 貢献サイトを記録するリスト
        contributing_sites = list()

        work_sites = list(sites)

        # Shuffle the order of sites, so that the order of sites in the given 
        # list does not affect sites selected among ties.
        # Remove comment marker of the next line to let it work.
        # random.shuffle(work_sites)

        # J: リスト中のサイトを順に、目標達成に必要なものかチェック
        # Each site in the list is checked if it is contributing to the 
        # achievement of the target
        for site in work_sites:
            val = self.get_value_of_feature_in_site(feature, site)

            if not val in needed_units: 
                # the value is too low to make contribution.
                continue
            if needed_units[val] == 0: 
                # A local unit of this value is no more needed.
                continue

            # J:貢献サイトリストに加える。
            # Add to the contributing site list.
            contributing_sites.append(site) 

            # J: また、そのサイトのvalue について、必要サイト数をひとつ減らす。
            # Number of local sites of this value needed to achieve the target
            # is decremented by 1. 
            # Thus, only required number of sites among ties are counted.
            needed_units[val] -= 1

            if len(contributing_sites) == target:
                break

        return contributing_sites
