# secsel
#
#     2018-06-16 by TAKENAKA, Akio
# rev 2020-12-05 by TAKENAKA, Akio
#
# rev 2021-06-26 by TAKENAKA, Akio  
#    Minor changes in handling errors in loading data files

import os
import sys
import re
import datetime

from . import targets
from . import local_units
from . import conflict_table
from . import conflict_solver
from . import secsel_core

from . import site_cost
from . import scatter_penalty

# J: ファイル入出力時の文字コード指定
chr_code = 'utf-8'  

# J:  全体のラッパーとなるクラス。
# J:  本クラス以外のクラスは、ユーザは意識する必要はない。

class SecSel():
    """
    Main class of SecSel system.
    Users work only with this class.
    
    To use this class, 
    1. create an object of this class calling site_selector.SiteSelector(),
    2. load parameters calling load_parameter_file(), 
    3. let it do the jop calling run().
    Result files will be created as directed in the parameter file.
    """

    def __init__(self):

        self.local_units = None             # LocalUnits object
        self.adjusted_local_units = None    # LocalUnits object, conflicts solved

        self.targets = None                 # Targets object

        self.conflicts  = None              # ConflictTable object
        self.solver  = None                 # ConflictSolver object

        self.site_cost = None               # SiteCost object


        self.scatter_penalty = None     #  ScatterPenalty object


        # For internal use

        self.features   = list()  # sorted list of all features
        self.req_n_list = list()  # list of required times of occurrence 
                                  # sorted in ascending order
                                  # Each element is uniquie.
                                  
        self.reserved_sites = []  # sites reserved prior to selection

        self.fresh = True         # The object is fresh until run() is called. 

        # Resuls of each run

        self.selected_sites = []  # selected sites

        # Keys will be feature ID, and the values are lists of site IDs.
        # The sites in a list contribute to the achievement of target of 
        # the corresponding feature.

        self.sites_for_feature = dict()


        # For stats of all runs

        # List of lists. 
        # Each list is a list of selected sites in individual runs.
        self.list_of_lists_of_selected_sites = list()
        
        # Keys will be site IDs and values, times selected
        self.selected_times = dict() 

        # Keys will be site IDs and the values are dictionaries.
        # The inner dictionaries' keys will be feature IDs, and values,
        # times the site is selected for individual features.
        
        self.selected_times_for_feature = dict()

        # List of costs of sites selected in all runs.
        # When there are pre-reserved sites, their costs are included.

        self.list_of_total_costs = list()


        # Parameters for sile selection are stored in a dictionary
        self.parameters = dict()  

        self.parameters['local_units_file'] =  ''    # LocalUnits data file
        self.parameters['default_top_n'] =  0        # default value of target
                                                     #  (number of top units to be covered)
        self.parameters['top_n_file'] =  ''          # target data file
        self.parameters['conflict_file'] = ''        # conflict data file
        self.parameters['pre_reserved_sites_file'] = ''  # Pre-reserved sites data file
        self.parameters['max_upset'] =  0            # Maximum upset of less effective but less
                                                     # costly/scatterd sites.  0 for no upset.

        self.parameters['consider_cost_first'] =  False    # True: cost considered first
                                                           # False(default): scatter penalty 
                                                           # considered first
        self.parameters['site_cost_file'] =   ''           # site cost data file
        self.parameters['ignore_cost'] = 0                 # Set 1 to ignore cost in site selection 
                                                           # (even when cost file is given)
        self.parameters['scatter_penalty_algorithm'] =  '' # algorithm of scatter penalty calculation
        self.parameters['scatter_penalty_file'] = ''       # scatter penalty data file

        self.parameters['output_file_base'] =   ''     # base name of output file
        self.parameters['save_each_run'] = True        # If true, result of each run will be saved.
        self.parameters['result_file_format'] = 'SITE' # Format of a result file of each run.
                                                       # [SITE, SITE_COST, FULL]
        self.parameters['n_run'] =   1                 # Repetion times. Default is 1.

        self.result_format_list = ['SITE', 'SITE_COST', 'FULL']

        # print("VERSION 1.0.4 (2010-12-06)")

    # J: #######################################
    # J: パラメータファイルの読み込み

    def load_parameter_file(self, file_name):
        """
        This function is called to load a parameter file.
        This function should be called before the execution of run()
        Execution is terminated if the parameter file contains invalid data.

        Parameter
        file: name of parameter file to load  (string)
        """

        if not os.path.isfile(file_name):
            print("!! parameter file", file_name, "does not exist.")
            print("!! Failed to load parameter file.")
            sys.exit()

        error_flag = False

        fin = open(file_name, 'rt', encoding = chr_code)

        while True:
            line = fin.readline()
            if not line:
                break

            line = re.sub('#.*$', '', line) #  '#' and what follows areremoved
            line = re.sub('\s+$', '', line) #  Spaces before the end of line are removed
            line = re.sub('^\s+', '', line) #  Spaces preceding the first non-space characters are removed

            if re.match('\s*$', line):
                continue             # blank line is skipped

            items = line.split()  

            parameter = items[0]   # Name of parameter
            parameter = re.sub('\s+$', '', parameter) # Remove tailing spaces.

            val = ""
            
            if len(items) >= 2:
                val  = items[1]        # Value to be assigned

            rvl = self.__check_and_set_parameter(parameter, val)

            if rvl == False:  # Check return value.
                error_flag = True

        fin.close()

        if error_flag == True:
            print("!! Failed to run due to parameter error")
            sys.exit()


    # J: ####################################
    # J: パラメータの値の取得のインタフェース
    # J: パラメータ名は case insensitive (内部的にはすべて小文字)
    
    def get_parameter(self, parameter_name):
        """
        This funcion is a getter of parameter value.
        
        Parameter
        parameter_name: string
            case insensitive. Internally, all parameter names are in lower case.
        Return
        value of the paramter. Type depends on the parameter.
             If invalid parameter_name is given, None is returned.
        """

        parameter_name = parameter_name.lower()
        
        if not parameter_name in self.parameters:
            return None
        else:
            return self.parameters[parameter_name]


    # J: ####################################
    # J: パラメータの値の設定のインタフェース

    def set_parameter(self, parameter_name, parameter_value):
        """
        This function is a user interface to set parameter value.
        
        If this function is called after load_parameter_file(),
        parameter values in the parameter file will be overridden
        All works are handed to __check_and_set_parameter().

        Parameter
        parameter_name: string
            case insensitive. Internally, all parameter names are in lower case.
        parameter_value: 
            value to be set. The type depends on the parameter.

        Return
        boolean:
            True if the name and value of the parameter are valid.
            False if either are not valid.
        """

        rvl = self.__check_and_set_parameter(parameter_name, parameter_value)

        return rvl


    # J: ####################################
    # J: サイト選択を実行。繰り返しが必要なら繰り返す

    def run(self):
        """
        This function executes site selection.
        Results will be stored in result files.
        """

        if self.fresh != True:
            print ("!! run() should be called only for fresh object.")
            sys.exit()

        if not self.parameters['output_file_base']:
            print("!! Output file name should be given.")

        print("# Output file base set as", self.parameters['output_file_base'])

        # Prepare objects

        self.__load_local_units()  # Required

        self.__load_targets() 
        self.__make_req_n_list()

        self.__load_conflict_table()  # Optional
        self.__load_reserved_sites()  # Optional

        self.__load_site_cost()       # Optional

        if self.parameters['scatter_penalty_algorithm']: # Scatter penalty is to be considered.

            rvl = self.__set_scatter_penalty_algorithm()
            if rvl == False:  # The given algorithm is not implemented
                print("!! Failed due to invalid scatter penalty algorithm")
                sys.exit()


        print("\n[", self.parameters['n_run'], "] runs:", end = "")

        # Run

        for i_run in range(self.parameters['n_run']):
            print(i_run + 1, end = " ")
            sys.stdout.flush()
            self.__run_a_run(i_run)

            self.list_of_lists_of_selected_sites.append(list(self.selected_sites))

            total_costs = self.__get_total_cost()
            self.list_of_total_costs.append(total_costs)

            # Save the results of this run.
            if self.parameters['save_each_run'] == True:
                result_file = self.__get_file_name_for_a_run(i_run)
                self.__save_results_of_a_run(result_file)


        # Save statistics 

        self.__save_stats()

        self.fresh = False  # No longer fresh

        print("\n---completed---")


    # J: ####################################
    # J: 各選択試行で選ばれたサイトのリストを、すべての試行について
    # J: 並べたリストを返す。
    # J: 事前保全サイトがある場合はこれも含む

    def get_list_of_lists_of_selected_sites(self):
        """
        # This function returns the list of lists.
        # Each list is a list of IDs of selected sites in individual runs.
        # If there are pre-reserved sites, they are included in the lists.

        Return
        list of lists of site IDs (string)
        """

        return self.list_of_lists_of_selected_sites


    # J: ####################################
    # J: これまでの選択（複数試行）で最終的に選ばれたサイトの総コストの
    # J: リストを返す。事前保全サイトがある場合はこれも含む

    def get_list_of_costs(self):
        """
        This function returns list of costs of sites selected in individual runs.
        When there are pre-reserved sites, their costs are included.

        Return
        list of costs of selected sites (float)
        """

        return  self.list_of_total_costs


    # J: ####################################
    # J: 指定回数目の試行結果を出力する結果ファイルの名前を返す。

    def get_file_name_for_a_run(self, i_run):
        """
        This function returns a file name to save results of a run.
        
        Parameter
        i_run: integer
            The serial number of the run (starts from 1)
            
        Return
        string: 
            File name made from output_file_base and i_run.
        """

        return  self.__get_file_name_for_a_run(i_run - 1)


    # J: ####################################
    # J: 集計結果ファイルの名前を返す。

    def get_summary_file_name(self):
        """
        This function returns the name of a summary file of multiple runs.
        
        Return
        string: 
            File name
        """

        return  self.__get_summary_file_name()


    ##########  Followings are prive methods used only within this class ##########

    # J: ####################################
    # J: １回のサイト選択を実行

    def __run_a_run(self, i_run):
        """
        Ths function executes i_run'th run of site selection.
        
        Parameter
        i_run: integer
            i_run is a serial number of this run (starting from 0). 
            The name of the result file of a run includes this number.
        """

        # Solve conflicts among features
        if not self.conflicts is None:
            # Solved LocalUnits object will be stored as self.adjusted_local_units
            self.__solve_conflicts()

        # Working LocalUnits object

        target_local_units = self.local_units

        if self.adjusted_local_units is not None:
            target_local_units = self.adjusted_local_units

        # List of selected sites. 
        # The sites are in the order of selection 
        # The first element of the list is the site selected first.
        pa_list = list() 

        # Pre-reserved sites will be the first to be stored in pa_list.
        if len(self.reserved_sites) > 0:
            pa_list = list(self.reserved_sites)  # make a deep copy

    #######################

        #  Starting from features with smaller targets 
        #  (number of top value local units covered)

        for req_n in self.req_n_list: # req_n_list is in ascending order.

            ss_core = secsel_core.SecSel_Core()

            # In the first repetition, pa_list is a list of 
            # pre-reserved sites (if any).
            # In the second and later repetiton, pa_list is a list of 
            # so-far selected sites.

            ss_core.set_reserved_sites(pa_list)  

            if (self.site_cost is not None) and (self.parameters['ignore_cost'] != 1): 
                ss_core.set_site_cost(self.site_cost)

            ss_core.set_scatter_penalty_obj(self.scatter_penalty)

            if self.parameters['consider_cost_first'] == True:
                ss_core.set_to_consider_cost_first(True)
            else:
                ss_core.set_to_consider_cost_first(False)

            ss_core.set_max_upset(self.parameters['max_upset'])


            # Deep copy of LocalUnits object
            work_local_units = target_local_units.dup()

            #  Get features of which target is req_n or more
            target_features = self.targets.get_features_requiring_n_or_more_sites(
                                                 self.features,
                                                 req_n)

            # Features not included in 'target_features' are no longer
            # need to be considered for conservation.

            dif_features = set(self.features).difference(set(target_features))


            #  Remove features which are no longer considered
            if len(dif_features) > 0: 
                work_local_units = work_local_units.dup_without_given_features(dif_features)

            ss_core.set_local_units(work_local_units)

            ss_core.set_top_n_upto(req_n)  # conservation target (number of top-value units)

            # Execution of site selection
            # Obtained list of selected sites, pa_list, will be the pre-reserved sites
            # in the next repetion of for loop with req_n. 
            # J: サイト選択の実行
            # J: 得られた選択サイト pa_list は、req_n の次の繰り返しで選択済みサイトとして登録

            pa_list = ss_core.select_sites() 

        # End of loop with maximum req_n.

        # Get results from SecSel object
        self.selected_sites = ss_core.get_selected_sites()

        # Evaluate contributions of sites selected in this run.
        self.__eval_contributions_of_selected_sites()


    # J: ####################################
    # J:  LocalUnits の情報を、指定したファイルから読み込む
    # J:  ファイルは、タブ区切りのテキストファイル。
    # J:  行が区画、列が多様性要素。１列目は区画のID、2列目からが各要素の評価値。

    def __load_local_units(self):
        """
        This function is called to load a data file for a LocalUnits object.
        The file name should be given via load_parameter_file() or
        set_parameter() before this function is called.
        The data file is a tab-separted text file.
        The first line is a header line. (e.g., Site[tab]Sp001[tab]Sp002...)
        Each column represents each feature.
        Each line starts with site ID followed by values of local units of
        individual features in the site.
        """

        local_units_file = self.parameters['local_units_file']

        if not local_units_file:
            print("!! Locat units file not given.")
            sys.exit()

        if not os.path.isfile(local_units_file):
            print("!! ", local_units_file,  "does not exist.")
            print ("!! Failed to load local units file.")
            sys.exit()

        self.local_units = local_units.LocalUnits()
        rvl = self.local_units.load_from_file(local_units_file)

        if (rvl == False) :
            print("!! Failed to load local units file.")
            sys.exit()


        print("# Local units file", local_units_file, "loaded.")

        self.features = self.local_units.get_features()
        self.features.sort()


    # J: ####################################
    # J: 必要保全区画数の設定をファイルから読み込む。
    # J: デフォルトの目標値が設定されておらず、ファイルも設定
    # J: されえいない場合は例外を発生する。

    def __load_targets(self):
        """
        This function is called to load a data file for a Target object.
        The file name should be given via load_parameter_file() or
        set_parameter() before this function is called.
        
        If default_top_n is not set and target file is not given,
        exectuion is terminated.
        
        Each line of the file consists of feature ID 
        """
        default_top_n = self.parameters['default_top_n']
        top_n_file = self.parameters['top_n_file']

        if default_top_n == 0 and top_n_file == '':
            print("!! Conservation targets not given: default value and/or target file is required.")
            sys.exit()

        self.targets = targets.Targets()

        self.targets.set_default_requirement(default_top_n)

        if top_n_file == '':
            return

        if not os.path.isfile(top_n_file):
            print("!! ", top_n_file,  "does not exist.")
            print("!! Failed to load top-n file.")
            sys.exit()

        rvl = self.targets.load_from_file(top_n_file)

        if (rvl == False) :
            print("!! Failed to load top-n file.")
            sys.exit()


        print("# Top N file", top_n_file, " loaded.")


    # J: ####################################
    # J:  各要素の必要出現数を調べ、その重複のないリストを作る（昇順）
    # J:  すでに self.local_units と self.targets に情報が読み込まれて
    # J:  いることが必要。

    def __make_req_n_list(self):

        """
        This function makes a list of required numbers of top-value local units
        of individual features. The result is stored in self.req_n_list.
        The list is used  in _run_a_run().
        """

        # A set of required numbers of top units of individual features.
        n_set = set() 

        for feature in self.features:
            n = self.targets.get_required_n_for_feature(feature)
            n_set.add(n)

        req_n_list = list(n_set)  # Elenents of a set are unique.
        req_n_list.sort()         # in ascending order

        self.req_n_list = req_n_list


    # J: ####################################
    # J: conflict 情報のファイルが設定されていたら、内容を読み込む
    # J: 
    # J:  要素群A と B が排他的な場合、１行目に要素群A の要素をタブ区切り、
    # J:  ２行目に要素群Bの要素を同じくタブ区切りで記載する。
    # J:  さらに要素群CとD が排他的であれば、A、B の行のあと、一行空白行を
    # J:  置いてから、C の要素の行、Dの要素の行と続ける。

    def __load_conflict_table(self):
        """
        This function is called to load a data file for a ConflictTable object.
        
        """

        conflict_file = self.parameters['conflict_file']

        if conflict_file == '':
            return

        if not os.path.isfile(conflict_file):
            print("!! ", conflict_file,  "does not exist.")
            print("!! Failed to load a conflict file.")
            sys.exit()

        self.conflicts = conflict_table.ConflictTable()
        rvl = self.conflicts.load_from_file(conflict_file)

        if (rvl == False) :
            print("!! Failed to load a conflict file.")
            sys.exit()


        print("# Conflict file", conflict_file, "loaded.")


    # J: ####################################
    # J: conflict 情報に基づいて、local_units の取捨選択

    def __solve_conflicts(self):
        """
        This funcion is called to solve conflicts among features.
        """

        if self.conflicts is None:
            return

        self.solver = conflict_solver.ConflictSolver()

        # J: local units オブジェクトのセット
        self.solver.set_local_units(self.local_units)    
        # J: 必要保全数の情報のセット
        self.solver.set_targets(self.targets)            
        # J: conflictの情報を読み込む
        self.solver.set_conflict_table(self.conflicts)   

        # Get LocatUnits object without conflicts
        self.adjusted_local_units = self.solver.solv_conflict()


    # J: ####################################
    # J: 事前保全サイトファイルが設定されていたら、内容を読み込む
    # J: ファイルはテキストファイル。一行に１サイトでも、複数サイトをタブ区切り
    # J: で並べてもよい。

    def __load_reserved_sites(self):
        """
        This function is called to load a data file of pre-reserved sites.
        Site IDs may be in multipe lines.
        Each line may contain more than one Site IDs separated by tabs.
        """

        reserved_sites_file = self.parameters['pre_reserved_sites_file']

        if reserved_sites_file == '':
            return

        if not os.path.isfile(reserved_sites_file):
            print("!! pre_reserved sites file", reserved_sites_file, "does not exist.")
            print("!! Failed to load a pre_reserved sites file.")
            sys.exit()

        self.reserved_sites = list() 

        fin = open(reserved_sites_file, 'rt', encoding = chr_code)

        line = fin.readline() # Header

        while True:
            line = fin.readline()
            if not line:
                break

            # A line of the file may consists of more than one site IDs 
            # separated by tabs
            sites = re.split('\s+', line[:-1])  

            self.reserved_sites.extend(sites)

        fin.close()

        print("# Pre_reserved sites file", reserved_sites_file, "loaded.")


    # J: ####################################
    # J:  SiteCost が使うサイトごとのコストのデータファイルが設定されていたら、
    # J:  内容を読み込む。
    # J:  各行に、サイトID とそのサイトのコストをタブ区切りで並べる

    def __load_site_cost(self):
        """
        This function is called to load a data file of costs to add individual sites.
        The data file contains lines, each of which consists of site ID and its cost.
        The ID and cost are separated by a tab.
        """

        site_cost_file = self.parameters['site_cost_file']

        if site_cost_file == '':
            return

        if not os.path.isfile(site_cost_file):
            print("!! site cost file", site_cost_file, "does not exist.")
            print("!! Failed to load a site cost file.")
            sys.exit()

        self.site_cost = site_cost.SiteCost()
        rvl = self.site_cost.load_cost_data(site_cost_file)

        if (rvl == False) :
            print("!! Failed to load a site cost file.")
            sys.exit()

        print("# Site cost file", site_cost_file, "loaded.")


    # J: ####################################
    # J: ScatterPenalty が使うアルゴリズムのセット

    def __set_scatter_penalty_algorithm(self):
        """
        This function is prepare implementation of scatter penalty algorithm.
        The algorithm used is cotrolled by the parameter scatter_penalty_algorithm.
        """

        algorithm = self.parameters['scatter_penalty_algorithm']
        
        # Scatter penalty not considered
        if algorithm == '':
            return

        self.scatter_penalty = scatter_penalty.ScatterPenalty()
        rvl = self.scatter_penalty.choose_algorithm(algorithm)

        if rvl == False:
            print("!! Algorithm of scatter penalty ", algorithm, 
                           " not valid.")
            sys.exit()

        scatter_penalty_file = self.parameters['scatter_penalty_file']
        
        if scatter_penalty_file == '':
            return

        rvl = self.scatter_penalty.load_data(scatter_penalty_file)
        if (rvl == False) :
            print("!! Failed to load a scatter penalty data file.")
            sys.exit()

        print("# Scatter Penalty file", scatter_penalty_file, "for", algorithm, "loaded.")


    # J: ####################################
    # J: 1回の試行試行で選ばれたサイトの貢献を評価

    def __eval_contributions_of_selected_sites(self):
        """
        This funciton evaluates contribution of selected sites to achive
        consevation targets of individual features. The results will be 
        stored in a dictionary self.selected_times_for_feature.
        
        Keys will be site IDs and the values are dictionaries.
        The inner dictionary's keys will be feature IDs, and values,
        times the site is selected for individual features.
        """

        # J: feature がキー、そのfeature のためのsites のリストが値
        # Keys will be feature IDs, and values will be lists of sites contributing
        # to the achievement of conservation targets of individual features.
        self.sites_for_feature = dict()

        target_local_units = self.local_units

        # If conflicts are solved, used solved LocalUnits object
        if self.adjusted_local_units is not None:
            target_local_units = self.adjusted_local_units

        # Pre-reserved sites and selected sites connected
        # all_selected_sites = self.reserved_sites + self.selected_sites

        # for site in all_selected_sites: 
        for site in self.selected_sites: # For each site selected.

            # For total stats.
            # Count times selected for each site.

            if site not in self.selected_times:
                self.selected_times[site] = 1
            else:
                self.selected_times[site] += 1

            if site not in self.selected_times_for_feature:
                self.selected_times_for_feature[site] = dict()


        # Make lists of sites contributing to the achievement of
        # conservation targets of indificual features.

        for feature in self.features:  # F
            required_n = self.targets.get_required_n_for_feature(feature)

            # Sites contributing to the achievement of conservation target
            # of this feature

            sites = target_local_units.get_contributing_sites(feature, 
                                        list(self.selected_sites), required_n)
            #                                  all_selected_sites, required_n)

            self.sites_for_feature[feature] = sites

            #  For total stats

            for site in sites:

                if site not in self.selected_times_for_feature:
                    self.selected_times_for_feature[site] = dict()

                if feature not in self.selected_times_for_feature[site]:
                    self.selected_times_for_feature[site][feature] = 1
                else:
                    self.selected_times_for_feature[site][feature] += 1


    # J: ####################################
    # J: １回の試行での選択結果をファイルに保存する。
    # J:
    # J: ヘッダ行に続いて、選ばれた順にサイトの情報が記録される。
    # J: 出力形式は、形式指定が SITE の場合
    # J: 最初がsite ID、次が選ばれた順番（1から。事前保全サイトがある場合は
    # J: それらの順番は0と表示。
    # J:   形式指定が SITE_COST および FULL の場合
    # J: 続いてそのサイトまでの積算コスト。コストのデータがない場合には0。
    # J: さらに、形式指定が FULL の場合、 そのサイトがどの要素の目的達成のために選択され
    # J: たのかの情報（1なら、その要素のため、0なら非該当）が全要素数だけ並び、さらに
    # J: そのあとに、そのサイト内での各要素の評価値が並ぶ。評価値はデータファイル
    # J: から読み込んだ情報だが、要素間で排他的な関係がある場合で、そのサイトでは
    # J: その要素の保全は行わない場合には-99 と記録される。

    def __save_results_of_a_run(self, result_file):
        """
        This function is called to save the result of a single run 
        of site selection. Format of the result file is controlled by
        the parameter 'result_file_format'.
        
        If the parameter is 'SITE', each line consists of site ID and order 
        of selection (starts from 1.  If there are pre-reserved sites, their
        orders are 0).
        If the parameter is 'SITE_COST' or 'FULL', order of selection is followed by 
        the cost of the site.
        If the parameter is 'FULL', contribution of the site to achieve target
        of individual features (1/0), and the value of local units of individual
        features follow the above. If the local unit is ignored as a result of
        solving conflicts among features, -99 is shown instead of the value of
        the local unit.
        
        Parameter
        result_file: name of a file to make (string)
        """

        fout = open(result_file, 'wt', encoding = chr_code) 

        # Header
        fout.write("site\torder")
        
        if not self.parameters['result_file_format'] == "SITE":
            # when SITE_COST or FULL
            fout.write("\ttotal_cost")

        if self.parameters['result_file_format'] == "FULL":

            for feature in self.features:
                # Contribution to the feature (1/0)
                fout.write("\tFor_%s" % feature)    

            for feature in self.features:
                # value of the local unit of this feature
                fout.write("\tVal_%s" % feature) 

        fout.write("\n")


        target_local_units = self.local_units

        if self.adjusted_local_units is not None:
            # Use conflict-solved LocalUnits object if it exists.
            target_local_units = self.adjusted_local_units

        # Pre-reserved sites and selected sites connected
        # all_selected_sites = self.reserved_sites + self.selected_sites

        """
        # Keys will be feature IDs and values, lists of sites
        # contributing to the achievement of target of individual features.
        sites_for_feature = dict()

        for feature in self.features:
            required_n = self.targets.get_required_n_for_feature(feature)
            sites_for_feature[feature] = target_local_units.get_contributing_sites(feature, 
                                             list(self.selected_sites), required_n)
        """
        
        cum_cost = 0    # Cumulative cost
        sel_order = 0   # Order of selection. 0 is assigned to pre-reserved sites.
        val = 0         # value of a local unit.


        for site in self.selected_sites: # For each site selected.

            if self.site_cost is not None:  # Cumulaive cost
                cost = self.site_cost.get_cost_of_site(site)
                cum_cost += cost 

            if site not in self.reserved_sites:  
                # If selected (not pre-reserved) site, increase order by 1
                sel_order = sel_order + 1

            # Site ID, order of selection
            fout.write("%s\t%d" % (site, sel_order))

            if not self.parameters['result_file_format'] == "SITE":
                # when SITE_COST or FULL
                fout.write("\t%f" % (cum_cost))


            if self.parameters['result_file_format'] == "FULL":
                # Contributes or not to this feature's target (1/0)
                for feature in self.features:
                # if site in sites_for_feature[feature]:
                    if site in self.sites_for_feature[feature]:
                        fout.write("\t1")
                    else:
                        fout.write("\t0")

                # value of the local unit of this feature
                # -99 if removed to solve conflist with other features

                del_features = None 
                if self.solver is not None:
                    del_features = self.solver.get_deleted_units_in_site(site)

                for feature in self.features:
                    val = self.local_units.get_value_of_feature_in_site(feature, site)
                    if (del_features is not None) and (feature in del_features):
                        val = -99

                    fout.write("\t%s" % (str(val)))

            fout.write("\n")

        fout.close()


    # J: #######################################
    # J: 全試行の選択の集計結果をファイルに保存する。
    # J: 選択回数の降順で、１行に１サイトの情報。
    # J:   サイトごとに選ばれた回数
    # J:   サイトごとに、各feature の目的達成のために選ばれた回数
    # J: result_file_format が FULLでない場合は、選択回数のみ出力。

    def __save_stats(self):
        """
        This function is called to make summary file of multiple runs.
        Each line represents the result of asite.
        Sites are arranged in descending order of times seleted.
        The format is controlled by the parameter result_file_format.
        If the parameter is 'SITE', the name of site and times selected are shown.
        If FULL, times the site was selected to contribute to individual features
        are also shown.
        """

        stats_file = self.__get_summary_file_name()
 
        fout = open(stats_file, 'wt', encoding = chr_code)

        # Header
        fout.write("site\ttimes") 

        if self.parameters['result_file_format'] == 'FULL':
            for feature in self.features:
                fout.write("\tFor_%s" % feature)

        fout.write("\n")

        # J: self.selected_times の site を、選択回数の値の降順でソート
        # J: 結果 は、キーと値のペア（タプル）が並んだリスト）
        # Sort tuples of site ID and selected times, in the order of
        # selected times (descending order).

        work = sorted(self.selected_times.items(), key = lambda x: x[1])

        sites_ordered = [x[0] for x in work]  # The first element of a tuple is a site ID
        sites_ordered.reverse()    # In descending order of times selected

        for site in sites_ordered:

            # Times selected
            fout.write("%s\t%d" % (site, self.selected_times[site]))

            # Times selected to achieve targets of individual features
            if self.parameters['result_file_format'] == 'FULL':
                for feature in self.features:
                    if feature in self.selected_times_for_feature[site]:
                        fout.write("\t%d" % (self.selected_times_for_feature[site][feature]))
                    else:
                        fout.write("\t0")

            fout.write("\n")

        fout.close()


    # J: ####################################
    # J: 選択サイトの総コストを返す。事前保全サイトも含む

    def __get_total_cost(self):
        """
        This function returns total cost of selected sites in a run.
        
        Return
        Float: Total cost of selected sites in a run

        """
        
        if self.site_cost is None: 
            return 0

        cum_cost = 0    # Cumulative cost

        for site in self.selected_sites: # For each site selected.
            cost = self.site_cost.get_cost_of_site(site)
            cum_cost += cost 

        return cum_cost


    # J: ####################################
    # J: パラメータの値の設定。内部用
    # J: パラメータ名と値の正当性をチェックして、正当であれば登録

    def __check_and_set_parameter(self, parameter, val):
        """
        This function is used to set parameter value.
        This function is called internally from set_parameter()
        and load_parameter_file().
        When either name or value of the parameter is invalid, 
        error message is printed to the standard output.
        
        Parameter
        parameter: sring
            Parameter name
        val: sring
            value of the parameter

        Return
        boolean:
            True if the name and value of the parameter are valid.
            False if either are not valid.
        
        """

        parameter = parameter.lower()  # in lower case

        parameter = re.sub('^\s+', '', parameter) #  Preceding spaces are removed
        parameter = re.sub('\s+$', '', parameter) #  Tailing spaces are removed

        error_flag = False

        if parameter == 'local_units_file':
            self.parameters['local_units_file'] = val

        elif parameter == 'default_top_n':
            if int(val) <= 0:
                print("default_top_n should be 1 or more")
                error_flag = True
            else:
                self.parameters['default_top_n'] = int(val)

        elif parameter == 'top_n_file':
            self.parameters['top_n_file'] = val

        elif parameter == 'pre_reserved_sites_file':
            self.parameters['pre_reserved_sites_file'] = val

        elif parameter == 'conflict_file':
            self.parameters['conflict_file'] = val

        elif parameter == 'save_each_run': 
            if int(val) == 1:
                self.parameters['save_each_run'] = True
            else:
                self.parameters['save_each_run'] = False

        elif parameter == 'result_file_format':
            if val != '':
                if not val in self.result_format_list:
                    print("Invalid result file format:", val)
                    error_flag = True
                else:
                    self.parameters['result_file_format'] = val

        elif parameter == 'n_run':
            if int(val) <= 0:
                print("Invalid n_run:", val)
                error_flag = True
            else:
                self.parameters['n_run'] = int(val)

        elif parameter == 'max_upset':
            if int(val)  < 0:
                print("Invalid max_upset:", val)
                error_flag = True
            else:
                self.parameters['max_upset'] = int(val)

        elif parameter == 'consider_cost_first':
            if int(val) == 1:
                self.parameters['consider_cost_first'] = True
            else:
                self.parameters['consider_cost_first'] = False

        elif parameter == 'site_cost_file':
            self.parameters['site_cost_file'] = val

        elif parameter == 'ignore_cost':
            self.parameters['ignore_cost'] = int(val)

        elif parameter == 'scatter_penalty_algorithm':
            if val != '':
                self.parameters['scatter_penalty_algorithm'] = val.upper()

        elif parameter == 'scatter_penalty_file':
            self.parameters['scatter_penalty_file'] = val

        elif parameter == 'output_file_base':
            self.parameters['output_file_base'] = val

        else:
            print("!! Invalid parameter name", parameter)
            error_flag = True

        if error_flag == True:
            return False  # Successful
        else:
            return True   # Not successful


    # J: ####################################
    # J: 指定回数目の試行結果を保存するファイルのファイル名を生成する。

    def __get_file_name_for_a_run(self, i_run):
        """
        This function makes a file name to save results of a run.
        
        Parameter
        i_run: integer
            The serial number of the run (starts from 0)
            
        Return
        string: 
            File name made from output_file_base and i_run.
            The serial number starts from 1.
        """

        x = self.parameters['n_run']
        digits = 0  # maximum digits required

        while (x >= 1):
            digits += 1
            x = x / 10

        format = "%%s-%%0%dd_PA.txt" % digits

        return format % (self.parameters['output_file_base'], (i_run + 1))


    # J: ####################################
    # J: 結果の集計ファイルのファイル名を生成する。

    def __get_summary_file_name(self):
        """
        This function returns the name of a summary file
        

        Return
        string: 
            Name of the summary file
        """

        stats_file = self.parameters['output_file_base'] + "_sum.txt"

        return stats_file
