#  Targets
#
#     2020-05-07  by TAKENAKA, Akio

import os.path
import sys
import re

# J: ファイル入出力時の文字コード指定
CHR_CODE = 'utf-8'  

# J: Top-N included を指標とする保護区選択システム SecSel のなかで、
# J: 各生物多様性要素それぞれについて top からいくつの local unit を
# J: 保全するかの目標を束ねて管理するクラス

class Targets:
    """
    This class handles conservation targets of individualfeatures to be protected.
    The conservation target for a feature is a number of sites with high priority 
    local units of the feature.
    """

    def __init__(self):

        # J: 個別の指定がない要素の目標保全サイト数
        # Default target.  Initial value is 0. Needs to be set larger value,
        # or targets for all features should be given in a target file
        # ('top_n_file' of the parameter file).
        self.default_n = 0

        # J: 辞書。 多様性要素がキー、目標保存サイト数が値。
        # The keys will be feature IDs and the values, target values (int)
        self.required_n_for_feature = {}   


    # J: ####################################
    # J: 個別指定がない要素の目標保全サイト数の設定
    def set_default_requirement(self, n):
        """
        This function sets default target. Unless otherwise set,
        targets of all features will be the default target.
        
        Parameter
        n: int
            Default target (number of sites with high priority local units)
        """
        self.default_n = n


    # J: ####################################
    # J: 要素ごとの必要保全 unit 数のデータをファイルから読み込む
    # J: 
    # J: self.load_from_file(file_name)
    # J:
    # J: データはタブ区切り。
    # J: 各行に、要素と目標保護サイト数が空白文字（含むタブ）区切りで書かれている。
    def load_from_file(self, file_name):
        """
        This function loads target info from a file.
        Each line of the file contains feature ID, and the 
        number of local units of the feature to be conserved.

        Parameters
        file_name: string
            Name of a file to load
            
        Return
        boolean
            True (if successful) or False (if failed)

        """

        if os.path.exists(file_name) is not True:
            print(file_name, " not found in ConservationTargets.load_from_file")
            return False

        fin = open(file_name, 'rt', encoding = CHR_CODE)

        line = fin.readline() # Header

        while True:

            line = fin.readline()
            if not line:
                break

            # feature ID and its target, separated by white space(s) incl. a tab.
            items = re.split('\s+', line[:-1])  

            feature  = items.pop(0)    # feature ID
            required_n = items.pop(0)  # target

            if (int(required_n) <= 0): # skip invalid target
                print("%s invalid data: %s -> %s" % (file_name, feature, required_n))
                continue

            self.required_n_for_feature[feature] = int(required_n)

        fin.close()

        return True


    # J:####################################
    # J: 指定要素の目標保全サイト数の取得
    def get_required_n_for_feature(self, feature):
        """
        This function returns the conservation target of a given feature.
        
        Parameters
        feature: string
            ID of a feature
        
        Return
        integer
            Conservation target of the feature.
        """

        if feature not in self.required_n_for_feature:
            if self.default_n <= 0:
                print("!! Output file name should be given.")
                sys.exit()

            return self.default_n
        else:
            return self.required_n_for_feature[feature]


    # J: ##################################
    # J: 与えられた要素のリスト中、目標保全サイト数が指定値以上のものを返す
    def get_features_requiring_n_or_more_sites(self, features, n):
        """
        This function returns a list of IDs of features of which 
        conservation target is 'n' or larger.
        
        Parameters
        features: list of strings
            List of feature IDs. The returned features are confined to 
            those in this list.
        n: integer
            Conservation target (number of sites with high priority local units)

        Return
        list of strings:
            List of feature IDs of which conservation target is 'n' or larger.
            The returned features are among those given as the parameter.
        """

        relevant_features = []

        for feature in features:
            if self.get_required_n_for_feature(feature) >= n:
                relevant_features.append(feature)

        return relevant_features
